package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.resources.ResourceLocation

class SkinGrid(
	val texture: ResourceLocation,
	val width: Float,
	val height: Float,
	val rows: Int = 16,
	val columns: Int = 16,
) {
	private var row = 0
	private var column = 0

	val imageWidth get() = width * columns
	val imageHeight get() = height * rows

	fun skip() {
		column++

		if (column >= columns) {
			jump()
		}
	}

	fun jump() {
		row++
		column = 0
	}

	fun next(): SkinElement {
		val element = SkinElement(texture, column * width, row * height, width, height, imageWidth, imageHeight)
		skip()
		return element
	}
}

fun ResourceLocation.element(
	x: Float,
	y: Float,
	w: Float,
	h: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, w, h, textureWidth, textureHeight)

class SkinElement @JvmOverloads constructor(
	val texture: ResourceLocation,
	val x: Float,
	val y: Float,
	val w: Float,
	val h: Float,
	val imageWidth: Float = 256f,
	val imageHeight: Float = 256f
) {
	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRaw(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) = render(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRawPartial(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) = renderPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderW(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1,
	) {
		render(stack, x, y, width = width, winding = winding)
	}

	@JvmOverloads
	fun renderW(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1,
	) = renderW(stack, x.toFloat(), y.toFloat(), width.toFloat(), winding)

	@JvmOverloads
	fun renderH(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1,
	) {
		render(stack, x, y, height = height, winding = winding)
	}

	@JvmOverloads
	fun renderH(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1,
	) = renderH(stack, x.toFloat(), y.toFloat(), height.toFloat(), winding)

	private val u0 = this.x / imageWidth
	private val v0 = this.y / imageHeight
	private val u1 = (this.x + w) / imageWidth
	private val v1 = (this.y + h) / imageHeight

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) {
		val winded = winding.translate(u0, v0, u1, v1)

		RenderHelper.drawTexturedRectUV(
			stack,
			x,
			y,
			width,
			height,
			winded.u0,
			winded.v0,
			winded.u1,
			winded.v1,
		)
	}

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) = renderRaw(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) {
		val u1 = (this.x + width) / imageWidth
		val v1 = (this.y + height) / imageHeight

		val winded = winding.translate(u0, v0, u1, v1)

		RenderHelper.drawTexturedRectUV(
			stack,
			x,
			y,
			width,
			height,
			winded.u0,
			winded.v0,
			winded.u1,
			winded.v1,
		)
	}

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) = renderRawPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)
}

val top_left_window_corner = SkinElement(RenderHelper.WIDGETS, x = 18f, y = 0f, w = 6f, h = 6f)
val top_right_window_corner = SkinElement(RenderHelper.WIDGETS, x = 24f, y = 0f, w = 6f, h = 6f)
val left_window_border = SkinElement(RenderHelper.WIDGETS, x = 18f, y = 4f, w = 3f, h = 5f)
val right_window_border = SkinElement(RenderHelper.WIDGETS, x = 27f, y = 3f, w = 3f, h = 5f)
val top_window_border = SkinElement(RenderHelper.WIDGETS, x = 22f, y = 0f, w = 5f, h = 3f)
val window_background = SkinElement(RenderHelper.WIDGETS, x = 30f, y = 12f, w = 6f, h = 6f)

val tab_right_connection = SkinElement(RenderHelper.WIDGETS, x = 30f, y = 0f, w = 3f, h = 5f)
val tab_left_connection = SkinElement(RenderHelper.WIDGETS, x = 33f, y = 0f, w = 3f, h = 5f)
val tab_background = SkinElement(RenderHelper.WIDGETS, x = 30f, y = 6f, w = 6f, h = 6f)
val bottom_left_window_corner = SkinElement(RenderHelper.WIDGETS, x = 18f, y = 6f, w = 6f, h = 6f)
val bottom_right_window_corner = SkinElement(RenderHelper.WIDGETS, x = 24f, y = 6f, w = 6f, h = 6f)
val bottom_window_border = SkinElement(RenderHelper.WIDGETS, x = 21f, y = 9f, w = 5f, h = 3f)
